require "xupdate/document"

module XUpdate
  #Represents XUpdate commands.
  class Commands
    attr_reader   :xupdate_document
    attr_accessor :target_document

    #xupdate:variable の記憶領域．

    #Storage of xupdate:variable
    @variable
    
    #コンストラクタ．
    # xupdate_document: target_document に対して適用する XUpdate 文書．
    # target_document: XUpdate を適用する文書．

    #Constructor
    # xupdate_document: A XUpdate document/command applies to @param target_document.
    # target_document: A Document applies from @param xupdate_document.
    def initialize(xupdate_document, target_document)
      @xupdate_document = xupdate_document
      @target_document = target_document
      @variable = {}
    end

    #@target_document に self を適用し，適用したデータを配列で返します．
    #@target_document は XUpdate::Document もしくは REXML::Element を選択できます．
    #XUpdate::Document の場合，すべてのコマンドを適用します．
    #REXML::Element の場合，自身をコマンドとみなして適用します．

    #Applies self to @param @target_document and returns patched data as array.
    #@param @target_document can be either a XUpdate::Document or an REXML::Element.
    #If a XUpdate::Document, applies all XUpdate commands in it.
    #If a REXML::Element, regards it is equal of XUpdate commands and applies it.
    def execute
      if @xupdate_document.kind_of? Document
        execute_document()
      else
        execute_command(@xupdate_document)
      end
    end
    
    private
    #@target_document に self を XUpdate 文書として適用し，適用したデータを配列で返します．

    #Applies self to @param @target_document and returns patched data as array.
    def execute_document
      result = []
      @xupdate_document.commands.each do |command|
        result << execute_command(command)
      end
      return result
    end
    
    #@target_document に self を XUpdate コマンドとして適用し，適用したデータを配列で返します．
    #適用したデータが 1 つの場合は，その要素のみを返します．

    #Applies self to @param @target_document and returns patched data as array.
    #If data is not multiple, returns data as the element.
    def execute_command(command)
      result = []
      if command.has_elements?
        command.to_a.each_with_index do |instruction,i|
          class << instruction;  attr_accessor :target;  end
          target_nodes = get_target(instruction.parent, instruction.target)
          if instruction.kind_of? REXML::Element
            target_nodes.each {|e| result << apply_xupdate_instructions(instruction, e) }
          else
            target_nodes.each {|e| result << apply_xml_element(instruction, e) unless e.nil? }
          end
        end
      else
        target_nodes = get_target(command)
        if command.name == "variable"
          result << apply_xupdate_variable(command)
        else
          target_nodes.each {|e| result << apply_pcdata(command, e) }
        end
      end
      result = result[0] if result.length == 1
      return result
    end

    #xupdate:instructions を適用し，適用したデータを返します．

    #Applies xupdate:instructions and returns patched data.
    def apply_xupdate_instructions(instruction, target)
      case instruction.name
      when "element"                 then apply_xupdate_element(instruction, target)
      when "attribute"               then apply_xupdate_attribute(instruction, target)
      when "text"                    then apply_xupdate_text(instruction, target)
      when "processing-instructions" then apply_xupdate_processing_instruction(instruction, target)
      when "comment"                 then apply_xupdate_comment(instruction, target)
      when "value-of"                then apply_xupdate_value_of(instruction, target)
      when "if"                      then puts "no-implementation: if"
      else                                puts "ignored instruction: " + instruction.name
      end
    end

    #PCDATA を適用し，適用したデータを返します．

    #Applies #PCDATA and returns patched data.
    def apply_pcdata(command, target)
      case command.name
      when "update"   then apply_pcdata_by_update(command, target)
      when "rename"   then apply_pcdata_by_rename(command, target)
      when "remove"   then apply_pcdata_by_remove(command, target)
      when "variable" then apply_pcdata_by_variable(command, target)
      else                 puts "igonored xupdate command (pcdata):\n" + command.to_s
      end
    end

    #PCDATA を xupdate:update で適用し，適用したデータを返します．

    #Applies #PCDATA by xupdate:update and returns patched data.
    def apply_pcdata_by_update(command, target)
      if target.kind_of? REXML::Text
        target.value = command.text
      elsif target.kind_of? REXML::Comment
        target.write(command.text)
      elsif target.kind_of? REXML::Attribute
        target.element.add_attribute(target.name, command.text)
      else
        puts "processing error (update): " + target.class.to_s
        puts command
      end
      return command
    end

    #PCDATA を xupdate:rename で適用し，適用したデータを返します．

    #Applies #PCDATA by xupdate:rename and returns patched data.
    def apply_pcdata_by_rename(command, target)
      if target.kind_of? REXML::Element
        target.name = command.text
      elsif target.kind_of? REXML::Attribute
        target.element.add_attribute(command.text, target.value)
        target.remove
      else
        puts "processing error (rename): " + target.class.to_s
        puts command
      end
      return command
    end

    #PCDATA を xupdate:remove で適用し，適用したデータを返します．

    #Applies #PCDATA by xupdate:remove and returns patched data.
    def apply_pcdata_by_remove(command, target)
      if target.kind_of? REXML::Element
        target.parent.delete(target)
      elsif [REXML::Attribute, REXML::Text, REXML::Comment].include?(target.class)
        target.remove
      else
        puts "processing error (remove): " + target.class.to_s
        puts command
      end
      return nil
    end

    #PCDATA を xupdate:variable で適用し，適用したデータを返します．

    #Applies #PCDATA by xupdate:variable and returns patched data.
    def apply_pcdata_by_variable(command, target)
      if target.kind_of? REXML::Element
        new_element = command.attributes["name"]
        target.add_element new_element
      else
        puts "processing error (variable): " + target.class.to_s
        puts command
      end
      return new_element
    end

    #XML 要素を適用し，適用したデータを返します．

    #Applies xml element and returns patched data.
    def apply_xml_element(instruction, target)
      if target.kind_of? REXML::Element
        if [REXML::Element, REXML::Text, REXML::Comment].include?(target.class)
          target.add instruction
        else
          puts "processing error (xml-element): " + target.class.to_s
        end
      elsif target.kind_of? REXML::Text
        if instruction.kind_of? REXML::Text
          target = instruction
        else
          puts "processing error (xml-element): " + target.class.to_s
        end
      else
        puts "igonored xupdate command (xml-element):\n" + instruction.to_s
      end
      return instruction
    end

    #xml 要素を適用し，適用したデータを返します．

    #Applies xupdate:element and returns patched data.
    def apply_xupdate_element(instruction, target)
      new_element = REXML::Element.new(instruction.attributes["name"])
      new_element.add_attribute(instruction.namespace["namespace"])
      add_node(instruction.parent, target, new_element)

      #下位ノードの適用．
      #applies lower layer.
      instruction.to_a.each do |child|
        if (child.kind_of? REXML::Element) && (child.namespace == "http://www.xmldb.org/xupdate")
          apply_xupdate_instructions(child, new_element)
        else
          new_element.add child
        end
      end
      return new_element
    end

    #xupdate:attribute を適用し，適用したデータを返します．

    #Applies xupdate:attribute and returns patched data.
    def apply_xupdate_attribute(instruction, target)
      new_attr = REXML::Attribute.new(instruction.attributes["name"], instruction.text)
      target.add_attribute new_attr
      target.add_attribute(instruction.namespace["namespace"])
      return new_attr
    end

    #xupdate:text を適用し，適用したデータを返します．

    #Applies xupdate:text and returns patched data.
    def apply_xupdate_text(instruction, target)
      new_text = REXML::Text.new(instruction.text)
      add_node(instruction.parent, target, new_text)
      return new_text
    end

    #xupdate:comment を適用し，適用したデータを返します．

    #Applies xupdate:comment and returns patched data.
    def apply_xupdate_comment(instruction, target)
      new_comment = REXML::Comment.new(instruction.text)
      add_node(instruction.parent, target, new_comment)
      return new_comment
    end
    
    #xupdate:processing-instruction を適用し，適用したデータを返します．

    #Applies xupdate:processing-instruction and returns patched data.
    def apply_xupdate_processing_instruction(instruction, target)
      new_instruction = Instruction.new(instruction.attributes["name"])
      add_node(instruction, target, new_instruction)
      return new_instruction
    end
    
    #xupdate:variable を適用し，適用したデータを返します．

    #Applies xupdate:variable and returns patched data.
    def apply_xupdate_variable(command)
      xpath = command.attributes["select"]
      data = REXML::XPath.first(@target_document, xpath).deep_clone
      @variable[command.attributes["name"]] = data
      return data
    end
    
    #xupdate:value-of を適用し，適用したデータを返します．

    #Applies xupdate:value-of and returns patched data.
    def apply_xupdate_value_of(instruction, target)
      name = instruction.attributes["select"]
      new_data = @variable[name[1..name.length]]
      add_node(instruction, target, new_data)
      return new_data
    end

    #コマンドの適用対象を求めます．

    #Gets targets to apply XUpdate commands.
    def get_target(command, target=nil)
      if target.nil?
        xpath = command.attributes["select"]
        result = REXML::XPath.match(@target_document, xpath)
      else
        result = [target]
      end
      return result
    end

    #xupdate:insert-before, xupdate:insert-after, xupdate:append を適用します．

    #Applies xupdate:insert-before, xupdate:insert-after, xupdate:append.
    def add_node(command, target, node)
      if command.name == "insert-before"
        #if [REXML::Element, REXML::Text, REXML::Comment].include?(target.class)
        target.previous_sibling = node
      elsif command.name == "insert-after"
        #if [REXML::Element, REXML::Text, REXML::Comment].include?(target.class)
        target.next_sibling = node
      elsif command.name == "append"
        #if [REXML::Element, REXML::Text, REXML::Comment].include?(target.class)
        child = command.attributes["child"]
        if child.nil? || child == "last()" || !target.has_elements?
          if node.kind_of? REXML::Element
            target.add_element node
          elsif node.kind_of? REXML::Text
            target.text = node
          else
            node.class.new(node, target)
          end
        else
          target.insert_before(target[child.to_i - 1], node)
        end
      else
        puts "igonored xupdate command (add node):\n" + target.to_s
      end
    end
  end
end
